#ifndef __LINUX_FINGERPRINTING_H
#define __LINUX_FINGERPRINTING_H

/*
 * include/linux/fingerprinting.h
 *
 * Jake Moilanen <moilanen@austin.ibm.com>
 * Copyright (C) 2006 IBM
 *
 * I/O Workload Fingerprinting
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
*/

#include <linux/types.h>
#include <linux/bio.h>

#define FP_TYPE_READ	0
#define FP_TYPE_WRITE	1
#define FP_PATTERN_SEQ	0
#define FP_PATTERN_RAND	1
#define FP_SIZE_SMALL	0
#define FP_SIZE_LARGE	1
#define FP_NUM_POINTS	(2 * 2 * 2)

struct fingerprint {
	__u8 type;
	__u8 pattern;
	__u8 size;
};

struct fp_snapshot {
	/* type */
	unsigned long reads;
	unsigned long writes;
	/* pattern */
	unsigned long head_pos;
	unsigned long avg_dist;
	/* size */
	unsigned long avg_size;
};

/* Number of reads/writes before classified as read */
#define FP_CLASS_READ_WRITE_RATIO	2

/* Number of sectors before pattern is random */
#define FP_CLASS_PATTERN_RAND		25

/* Number of sectors before size is large */
#define FP_CLASS_SIZE_LARGE		8

extern void update_fp_snapshot(struct bio * bio);
extern void calc_fp(struct fingerprint * fp, struct fp_snapshot * fp_ss);
extern void reset_fp_snapshot(struct fp_snapshot * ss);
extern void reset_fp(struct fingerprint * fp);
extern void consolidate_fp_snapshot(struct fp_snapshot * master, struct fp_snapshot * instance);
extern int fingerprint_state_show(struct seq_file *s, void *unused);
extern int fingerprint_snapshot_show(struct seq_file *s, void *unused);
extern int fingerprint_top_fitness_show(struct seq_file *s, void *unused);

/* XXX do this more intelligently */
#ifndef DIVLL_OP
#define DIVLL_OP
#if BITS_PER_LONG >= 64

static inline void divll(long long *n, long div, long *rem)
{
        *rem = *n % div;
        *n /= div;
}

#else

static inline void divl(int32_t high, int32_t low,
                        int32_t div,
                        int32_t *q, int32_t *r)
{
        int64_t n = (u_int64_t)high << 32 | low;
        int64_t d = (u_int64_t)div << 31;
        int32_t q1 = 0;
        int c = 32;
        while (n > 0xffffffff) {
                q1 <<= 1;
                if (n >= d) {
                        n -= d;
                        q1 |= 1;
                }
                d >>= 1;
                c--;
        }
        q1 <<= c;
        if (n) {
                low = n;
                *q = q1 | (low / div);
                *r = low % div;
        } else {
                *r = 0;
                *q = q1;
        }
        return;
}

static inline void divll(long long *n, long div, long *rem)
{
        int32_t low, high;
        low = *n & 0xffffffff;
        high = *n >> 32;
        if (high) {
                int32_t high1 = high % div;
                int32_t low1 = low;
                high /= div;
                divl(high1, low1, div, &low, (int32_t *)rem);
                *n = (int64_t)high << 32 | low;
        } else {
                *n = low / div;
                *rem = low % div;
        }
}
#endif

#endif /* #ifndef divll */

#endif /* __LINUX_FINGERPRINTINT_H */
