#ifndef __LINUX_GENETIC_H
#define __LINUX_GENETIC_H
/*
 * include/linux/genetic.h
 *
 * Jake Moilanen <moilanen@austin.ibm.com>
 * Copyright (C) 2004 IBM
 *
 * Genetic algorithm library
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. 
 */

#include <linux/list.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/seq_file.h>


#define GENETIC_HISTORY_SIZE		0x8
#define GENETIC_HISTORY_MASK		(GENETIC_HISTORY_SIZE - 1)

/* percentage of total number genes to mutate */
#define GENETIC_DEFAULT_MUTATION_RATE	15

/* XXX TODO Make this an adjustable runtime variable */
/* Percentage that an iteration can jump within the range */
#define GENETIC_ITERATIVE_MUTATION_RANGE 20

/* the rate that GENETIC_DEFAULT_MUTATION_RATE itself can change */
#define GENETIC_DEFAULT_MUTATION_RATE_CHANGE 4
#define GENETIC_MAX_MUTATION_RATE	45
#define GENETIC_MIN_MUTATION_RATE	10

#define GENETIC_DEBUG			0

#ifdef CONFIG_FINGERPRINTING
#define FP_DECAY			90
#define GENETIC_NUM_DEBUG_POINTS	5
#else
#define GENETIC_NUM_DEBUG_POINTS	4
#endif

#define GENETIC_PRINT_DEBUG		0
#define gen_dbg(format, arg...) do { if (GENETIC_PRINT_DEBUG) printk(KERN_EMERG __FILE__ ": " format "\n" , ## arg); } while (0)
#define gen_trc(format, arg...) do { if (GENETIC_PRINT_DEBUG) printk(KERN_EMERG __FILE__ ":%s:%d\n" , __FUNCTION__, __LINE__); } while (0)

struct gene_param_s;
struct genetic_s;
struct phenotype_s;

struct genetic_child_s {
	struct list_head	list;
	long long		fitness;
	unsigned long		num_genes;
	void			*genes;
	struct gene_param_s	*gene_param;
	void			*stats_snapshot;
	int			id;
};

typedef struct genetic_child_s genetic_child_t;

/* Here's a generic idea of what it the genes could look like */
struct gene_param_s {
	char 		*name;
	unsigned long	min;
	unsigned long	max;
	unsigned long	initial;
	void		(*mutate_gene)(genetic_child_t *, unsigned long);
};

typedef struct gene_param_s gene_param_t;

struct phenotype_s {
	struct list_head	phenotype;

	struct list_head	children_queue[2];
	struct list_head	*run_queue;
	struct list_head	*finished_queue;
	struct genetic_ops	*ops;

	char			*name;

	struct genetic_s	*genetic;		/* point back
							 * to genetic
							 * struct
							 */

	unsigned long		num_children;		  /* Must be power of 2 */
	unsigned long		natural_selection_cutoff; /* How many children
							   * will survive
							   */
	void			*stats_snapshot;
	unsigned long		child_number;

	/* percentage of total number of genes to mutate */
	long			mutation_rate;
	unsigned long		num_mutations;
	unsigned long		num_genes;

	genetic_child_t		**child_ranking;

	void			(*natural_selection)(struct phenotype_s *);

	/* This UID is bitmap comprised of other phenotypes that contribute
	   to the genes */
	unsigned long		uid;

	/* performance metrics */
	long long		avg_fitness;
	long long		last_gen_avg_fitness;

	unsigned long		fitness_history_index;
	long long		fitness_history[GENETIC_HISTORY_SIZE];

#if GENETIC_DEBUG
	unsigned long		debug_size;	/* number of longs in
						   debug history */
	unsigned long		debug_index;
	long long		*debug_history;
#endif
#ifdef CONFIG_FINGERPRINTING
	struct dentry 		*fp_dir;
	struct fingerprint	*fp;
	struct fp_snapshot	*fp_ss;
	unsigned long		***top_child;
	long long		***top_fitness;
	int			last_fingerprint;
#else
	long long		top_fitness;
#endif

	long long		from_top;

};

typedef struct phenotype_s phenotype_t;

/**
 * struct genetic_s - contains all data structures for a genetic plugin
 * @name: string that will identify this genetic alg. in debugfs and printk
 * @phenotype: list of all registered phenotypes
 * @child_number: the running child index (< @num_children)
 * @child_life_time: time in ms each child is ran before being swapped
 * @num_children: number of children in each generation (must be a power of 2)
 * @generation_number: increased once every child in a generation is ran
 * @defaults: when 1 the genetic library will hold all genes at defaults
 * @fingerprinting: when 1 the genetic library wil use gene fingerprinting if CONFIG_FINGERPRINTING
 */
struct genetic_s {
	char			*name;
	struct timer_list	timer;

	struct list_head	phenotype;

	unsigned long		child_number;
	unsigned long		child_life_time;
	unsigned long		num_children;

	unsigned long		generation_number;
	int 			defaults;

	/* private */
	struct dentry		*dir;
	struct dentry 		*phenotypes_dir;
	struct dentry 		*fingerprinting_dir;
#ifdef CONFIG_FINGERPRINTING
	int			fingerprinting;
#endif
};

typedef struct genetic_s genetic_t;

struct genetic_ops {
	void			(*create_child)(genetic_child_t *);
	void			(*set_child_genes)(void *);
	void			(*calc_fitness)(genetic_child_t *);
	void			(*combine_genes)(genetic_child_t *, genetic_child_t *,
						 genetic_child_t *);
	void			(*mutate_child)(genetic_child_t *);
	void			(*calc_post_fitness)(phenotype_t *); /* Fitness routine used when
								      * need to take into account
								      * other phenotype fitness
								      * results after they ran
								      */
	void			(*take_snapshot)(phenotype_t *);
	void			(*shift_mutation_rate)(phenotype_t *);
	int			(*gene_show)(struct seq_file *, void *);
#ifdef CONFIG_FINGERPRINTING
	void			(*get_fingerprint)(phenotype_t *);
	void			(*update_fingerprint)(phenotype_t *);
	void *			(*create_top_genes)(phenotype_t *);
	int			(*top_fitness_show)(struct seq_file *, void *);
	int			(*snapshot_show)(struct seq_file *, void *);
	int			(*state_show)(struct seq_file *, void *);
#endif
};

/* Setup routines */
int __init genetic_init(genetic_t ** in_genetic, unsigned long num_children,
			unsigned long child_life_time, int fingerprinting,
			char * name);
int __init genetic_register_phenotype(genetic_t * genetic, struct genetic_ops * ops,
				      unsigned long num_children, char * name,
				      unsigned long num_genes, unsigned long uid);
void __init genetic_start(genetic_t * genetic);

/* Generic helper functions */
void genetic_generic_mutate_child(genetic_child_t * child);
void genetic_generic_iterative_mutate_gene(genetic_child_t * child, unsigned long gene_num);
void genetic_generic_combine_genes(genetic_child_t * parent_a,
				   genetic_child_t * parent_b,
				   genetic_child_t * child);
void genetic_create_child_spread(genetic_child_t * child, unsigned long num_children);
void genetic_create_child_defaults(genetic_child_t * child);
void genetic_general_shift_mutation_rate(phenotype_t * in_pt);
int genetic_generic_gene_show(struct seq_file *s, void *unused);


/* XXX do this more intelligently */
#ifndef DIVLL_OP
#define DIVLL_OP
#if BITS_PER_LONG >= 64

static inline void divll(long long *n, long div, long *rem)
{
        *rem = *n % div;
        *n /= div;
}

#else

static inline void divl(int32_t high, int32_t low,
                        int32_t div,
                        int32_t *q, int32_t *r)
{
        int64_t n = (u_int64_t)high << 32 | low;
        int64_t d = (u_int64_t)div << 31;
        int32_t q1 = 0;
        int c = 32;
        while (n > 0xffffffff) {
                q1 <<= 1;
                if (n >= d) {
                        n -= d;
                        q1 |= 1;
                }
                d >>= 1;
                c--;
        }
        q1 <<= c;
        if (n) {
                low = n;
                *q = q1 | (low / div);
                *r = low % div;
        } else {
                *r = 0;
                *q = q1;
        }
        return;
}

static inline void divll(long long *n, long div, long *rem)
{
        int32_t low, high;
        low = *n & 0xffffffff;
        high = *n >> 32;
        if (high) {
                int32_t high1 = high % div;
                int32_t low1 = low;
                high /= div;
                divl(high1, low1, div, &low, (int32_t *)rem);
                *n = (int64_t)high << 32 | low;
        } else {
                *n = low / div;
                *rem = low % div;
        }
}
#endif

#endif /* #ifndef divll */

#endif
